/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

/* The following description is used by demos/make-demo-list.pl
 * to compile a list of all tests. */
/* DESCRIPTION
 * Echo test.
 *
 * Function: demo_echo1
 */


#include <gtk/gtk.h>
#include <gdam/gdambuffer.h>
#include "../gdamchannel.h"

typedef struct _DemoInfo {
	GdamChannel* channel;
	GtkWindow* top_level;
	GtkLabel* filename;
	GtkFileSelection* file_selection;

	GtkToggleButton* reverb_button;

	/* We keep a reference to these: */
	GtkAdjustment* num_echoes;
	GtkAdjustment* per_echo_fade;
	GtkAdjustment* per_echo_delay;
	GtkAdjustment* start_gain;

	guint mixer_id;
	guint echo_id;
	gboolean making_request;
} DemoInfo;

static void handle_toggle_button(DemoInfo* info)
{
	gboolean reverb;
	g_return_if_fail(info->echo_id != 0);
	reverb = gtk_toggle_button_get_active(info->reverb_button);
	gdam_channel_source_set_arg(info->channel,
                                    info->echo_id,
				    "reverb",
				    GDAM_ARG_TYPE_INT,
				    reverb);
}

static void set_echo_array(DemoInfo* info)
{
	int num_echoes = (int)(info->num_echoes->value);
	gfloat fade = (info->per_echo_fade->value);
	gfloat gain = (info->start_gain->value);
	gfloat delay = (info->per_echo_delay->value);
	int i;
	int* echoes;
	gfloat cur_delay = 0;
	//gboolean reverb = gtk_toggle_button_get_active(info->reverb_button);

	g_warning("info->start_gain: value=%f, upper/lower=%f/%f",
		info->start_gain->value,
		info->start_gain->lower,
		info->start_gain->upper);

	echoes = (int*) alloca(sizeof(int) * 2 * num_echoes);
	for (i = 0; i < num_echoes; i++) {
		g_message("echo %d: volume %f, delay %f", i,gain,cur_delay);
		echoes[2*i+0] = (int)((float)(1<<16) * gain);   /* volume */
		echoes[2*i+1] = (int)(44100.0 * cur_delay);     /* delay */
		cur_delay += delay;
		gain *= fade;
	}
	gdam_channel_source_set_arg(info->channel,
	                            info->echo_id,
				    "echoes",
				    GDAM_ARG_TYPE_INT_ARRAY,
				    num_echoes * 2,
				    echoes);
	gdam_debug("set %d echoes with delay %f, decay %f",
		num_echoes, delay, fade);
}



static void chose_file(GtkButton* button, DemoInfo* info)
{
	char* filename;
	g_return_if_fail(info->file_selection != NULL);
	filename = gtk_file_selection_get_filename(info->file_selection);
	gtk_label_set_text(info->filename, filename);
}

static void filesel_button_clicked(GtkButton* button, DemoInfo* info)
{
	if (info->file_selection) {
		GtkObject* to_kill = GTK_OBJECT(info->file_selection);
		info->file_selection = NULL;
		gtk_object_destroy(to_kill);
	} else {
		GtkWidget* filesel = gtk_file_selection_new("pick your mp3");
		GtkFileSelection* file_select = (GtkFileSelection*)filesel;
		info->file_selection = file_select;
		gtk_signal_connect(GTK_OBJECT(file_select->ok_button),
		                   "clicked",
		                   (GtkSignalFunc) chose_file,
				   info);
		/*gtk_signal_connect_object(
		                   GTK_OBJECT(file_select->cancel_button),
		                   "clicked",
				   (GtkSignalFunc) gtk_widget_destroy,
				   GTK_OBJECT(file_select));*/
		gtk_widget_show(GTK_WIDGET(file_select));
	}
}

static void play_callback(GtkButton* button, DemoInfo* info)
{
	gchar* filename;
	int mp3_id;
	if (info->echo_id != 0 || info->making_request)
		return;

	info->making_request = TRUE;
	gtk_label_get(info->filename, &filename);
	gdam_debug("create echo %s", filename);
	mp3_id = gdam_channel_create_source_generic(
			info->channel,
			"GdamMp3",
			1,
			"filename",
			GDAM_ARG_TYPE_STRING,
			filename);

	info->echo_id = gdam_channel_create_source_generic(
			info->channel,
			"GdamEcho",
			2,
			"echoes",
			GDAM_ARG_TYPE_INT_ARRAY,
			0, NULL,
			"filtered",
			GDAM_ARG_TYPE_INT,
			mp3_id);

	set_echo_array(info);
	if (!gdam_channel_mix_add(info->channel,
	                          info->mixer_id,
				  info->echo_id,
				  -1))
		g_warning("gain_add failed");
	info->making_request = FALSE;
}

#if 0

GdamSoundSpec* gdam_sound_spec_echo_v (GdamSoundSpec* sub,
                                       int            num_echoes,
	                               const GdamSoundSpecEchoElement* echoes)
{
	gdam_debug("gdam_channel_mix_add being called...");
	if (!gdam_channel_mix_add(info->channel,
	                          info->mixer_id,
				  info->echo_id,
				  -1))
		g_warning("echo_add failed");
}
#endif

static void stop_callback(GtkButton* button, DemoInfo* info)
{
	if (info->echo_id == 0 || info->making_request)
		return;
	if (!gdam_channel_detach_source(info->channel, info->echo_id))
		g_warning("detach source failed");
	info->echo_id = 0;
}


/* Helper + macro to cut down repetition in demo1_init(). */
static GtkWidget* _generic_button(const char* label, GtkSignalFunc callback,
                                 gpointer user_data)
{
	GtkWidget* button = gtk_button_new_with_label(label);
	gtk_signal_connect(GTK_OBJECT(button), "clicked", callback, user_data);
	gtk_widget_show(GTK_WIDGET(button));
	return button;
}

#if 0
static void echo_changed(GtkAdjustment* adjustment, DemoInfo* info)
{
	if (info->echo_id == 0)
		return;
	if (!gdam_channel_set_echoes(info->channel,
	                             info->echo_id, 
				     (int)(adjustment->value * 65536.0)))
		return;
}
#endif

static GtkWidget* show(GtkWidget* widget)
{
	gtk_widget_show(widget);
	return widget;
}
static GtkWidget* label(const char* text)
{
	return show(gtk_label_new(text));
}
#define generic_button(label, callback)				\
		_generic_button(label, GTK_SIGNAL_FUNC(callback), info)

void hello_world(void) { g_message("hello world"); }
void demo_echo1(GdamChannel*           channel,
                guint                  mixer_id)
{
	DemoInfo* info = g_new(DemoInfo, 1);
	GtkWidget* hbox;
	GtkWidget* vbox;
	//GtkAdjustment* adjustment;
	//GtkWidget* scroll;

	g_message("demo_info=%p",info);

	vbox = gtk_vbox_new(FALSE, 0);
	hbox = gtk_hbox_new(FALSE, 0);

	info->channel = channel;
	info->top_level = (GtkWindow*) gtk_window_new(GTK_WINDOW_TOPLEVEL);
	info->filename = (GtkLabel*) gtk_label_new("???");
	info->file_selection = NULL;
	info->making_request = FALSE;
	info->mixer_id = mixer_id;
	info->echo_id = 0;
	info->num_echoes = (GtkAdjustment*)gtk_adjustment_new(1,1,20, 1,1,1);
	info->per_echo_fade = (GtkAdjustment*)gtk_adjustment_new(0.5, 
	                                                         0.0, 1.0,
								 0.01, 0.1,
								 0.1);
	info->per_echo_delay = (GtkAdjustment*)gtk_adjustment_new(1.0,
	                                                         0.0, 10.0,
								 0.1, 0.1,
								 1.0);
	info->start_gain = (GtkAdjustment*)gtk_adjustment_new(0.5, 
	                                                         0.0, 1.0,
								 0.01, 0.1,
								 0.1);
	gtk_object_ref(GTK_OBJECT(info->num_echoes));
	gtk_object_ref(GTK_OBJECT(info->per_echo_fade));
	gtk_object_ref(GTK_OBJECT(info->per_echo_delay));
	gtk_object_ref(GTK_OBJECT(info->start_gain));

	gtk_widget_show(GTK_WIDGET(info->filename));
	gtk_box_pack_start_defaults(GTK_BOX(hbox), GTK_WIDGET(info->filename));
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                  generic_button("File...", filesel_button_clicked));
	gtk_widget_show(hbox);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), GTK_WIDGET(hbox));


	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                        generic_button("Play", play_callback));
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                        generic_button("Stop", stop_callback));
	gtk_widget_show(hbox);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), GTK_WIDGET(hbox));



	{
		GtkBox* inner = GTK_BOX(gtk_vbox_new(TRUE, 1));
		GtkBox* inner_hbox;
		inner_hbox = GTK_BOX(gtk_hbox_new(FALSE, 0));
		info->reverb_button = (GtkToggleButton*)
					gtk_toggle_button_new_with_label(
						"Reverb"
					);
		gtk_signal_connect_object_after(
					GTK_OBJECT(info->reverb_button),
					"toggled",
					GTK_SIGNAL_FUNC(handle_toggle_button),
					(GtkObject*)info);
		gtk_box_pack_start_defaults(inner_hbox,
				    show(GTK_WIDGET(info->reverb_button)));
		gtk_box_pack_start_defaults(inner_hbox, label("Num Echoes:"));
		gtk_box_pack_start_defaults(inner_hbox,
			    show(gtk_spin_button_new(info->num_echoes,
								     0.3, 1)));
		gtk_box_pack_start_defaults(inner,
		                            show(GTK_WIDGET(inner_hbox)));
		

		inner_hbox = GTK_BOX(gtk_hbox_new(FALSE, 0));
		gtk_box_pack_start_defaults(inner_hbox, label("Start Gain"));
		gtk_box_pack_start_defaults(inner_hbox,
			    show(gtk_hscrollbar_new(info->start_gain)));
		gtk_box_pack_start_defaults(inner,
		                            show(GTK_WIDGET(inner_hbox)));

		inner_hbox = GTK_BOX(gtk_hbox_new(FALSE, 0));
		gtk_box_pack_start_defaults(inner_hbox, label("Delay"));
		gtk_box_pack_start_defaults(inner_hbox,
			    show(gtk_hscrollbar_new(info->per_echo_delay)));
		gtk_box_pack_start_defaults(inner,
		                            show(GTK_WIDGET(inner_hbox)));

		inner_hbox = GTK_BOX(gtk_hbox_new(FALSE, 0));
		gtk_box_pack_start_defaults(inner_hbox, label("Decay"));
		gtk_box_pack_start_defaults(inner_hbox,
			    show(gtk_hscrollbar_new(info->per_echo_fade)));
		gtk_box_pack_start_defaults(inner,
		                            show(GTK_WIDGET(inner_hbox)));
		gtk_box_pack_start_defaults(GTK_BOX(vbox),
		                            show(GTK_WIDGET(inner)));
	}

	/* Callbacks... */
	gtk_signal_connect_object_after(GTK_OBJECT(info->start_gain),
				  "value_changed",
	                          GTK_SIGNAL_FUNC(set_echo_array),
				  (GtkObject*)info);
	gtk_signal_connect(GTK_OBJECT(info->start_gain), "value_changed",
	                    GTK_SIGNAL_FUNC(hello_world), NULL);
	gtk_signal_connect_object_after(GTK_OBJECT(info->per_echo_delay),
				  "value_changed",
	                          GTK_SIGNAL_FUNC(set_echo_array),
				  (GtkObject*)info);
	gtk_signal_connect_object_after(GTK_OBJECT(info->per_echo_fade),
				  "value_changed",
	                          GTK_SIGNAL_FUNC(set_echo_array),
				  (GtkObject*)info);
	gtk_signal_connect_object_after(GTK_OBJECT(info->num_echoes),
				  "value_changed",
	                          GTK_SIGNAL_FUNC(set_echo_array),
				  (GtkObject*)info);

#if 0
	adjustment = GTK_ADJUSTMENT(gtk_adjustment_new(
					1.0, 0.0, 3.0,	/* default, min, max */
					0.01, 0.1, 1.0));
	gtk_signal_connect(GTK_OBJECT(adjustment),
	                   "value_changed",
			   GTK_SIGNAL_FUNC(echo_changed),
			   info);
	scroll = gtk_hscrollbar_new(adjustment);
	gtk_range_set_update_policy(GTK_RANGE(scroll), GTK_UPDATE_CONTINUOUS);
	gtk_widget_show(scroll);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), GTK_WIDGET(scroll));
#endif

	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(info->top_level), GTK_WIDGET(vbox));
	gtk_widget_show(GTK_WIDGET(info->top_level));
}
