/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#ifndef __GDAM_MIDI_DEVICE_H_
#define __GDAM_MIDI_DEVICE_H_


typedef struct _GdamMidiDevice GdamMidiDevice;
typedef struct _GdamMidiDeviceClass GdamMidiDeviceClass;
typedef struct _GdamMidiNote GdamMidiNote;
typedef struct _GdamMidiPitch GdamMidiPitch;
typedef struct _GdamMidiParam GdamMidiParam;
typedef struct _GdamMidiSystemExclusive GdamMidiSystemExclusive;
typedef struct _GdamMidiTrap GdamMidiTrap;

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include <gtk/gtkobject.h>

typedef union _GdamMidiEvent GdamMidiEvent;

GtkType gdam_midi_device_get_type();
#define GDAM_TYPE_MIDI_DEVICE			(gdam_midi_device_get_type ())
#define GDAM_MIDI_DEVICE(obj)              (GTK_CHECK_CAST ((obj), GDAM_TYPE_MIDI_DEVICE, GdamMidiDevice))
#define GDAM_MIDI_DEVICE_CLASS(klass)      (GTK_CHECK_CLASS_CAST ((klass), GDAM_TYPE_MIDI_DEVICE, GdamMidiDeviceClass))
#define GDAM_IS_MIDI_DEVICE(obj)           (GTK_CHECK_TYPE ((obj), GDAM_TYPE_MIDI_DEVICE))
#define GDAM_IS_MIDI_DEVICE_CLASS(klass)   (GTK_CHECK_CLASS_TYPE ((klass), GDAM_TYPE_MIDI_DEVICE))

/* A function you can register to receive a particular subsets
 * of the midi event stream.  We call this a ``hook'';  you
 * make a hook with ``gdam_midi_device_connect''. */
typedef  void     (*GdamMidiHookFunc)    (GdamMidiDevice*    midi_dev,
	                                  GdamMidiEvent*     midi_event,
				          gpointer           user_data);

/* A function you can use to iterate over all known traps. (used by ui code) */
typedef  void     (*GdamMidiTrapIterator)(GdamMidiTrap*      trap,
                                          gpointer           iter_data);


typedef enum _GdamMidiEventType GdamMidiEventType;
enum _GdamMidiEventType {
	GDAM_MIDI_NOTE_ON,
	GDAM_MIDI_NOTE_OFF,
	GDAM_MIDI_PITCH_WHEEL,
	GDAM_MIDI_PARAMETER,
	GDAM_MIDI_SYSEX
};

/* For use with midi piano keys. */
struct _GdamMidiNote {
	GdamMidiEventType	type;
	int			instrument;	/* was ``instrument'' */
	int			velocity;
	int			midi_index;
	const char*		note_name;
	int			octave;
	gdouble			hertz;
};

/* For use with pitch wheel. */
struct _GdamMidiPitch {
	GdamMidiEventType	type;
	int 			channel;
	int			pitch;

	gdouble			norm_value;
};

/* For use with "data entry" aka "parameter" midi events. */
struct _GdamMidiParam {
	GdamMidiEventType	type;
	int 			channel;
	int			parameter;
	int			value;
	gdouble			norm_value;
};

struct _GdamMidiSystemExclusive {
	GdamMidiEventType	type;
	const char*		make_name;
	const char*		model_name;
	int			message_length;
	const guint8*		message;
};

union _GdamMidiEvent {
	GdamMidiEventType	type;
	GdamMidiParam		param_event;
	GdamMidiPitch		pitch_event;
	GdamMidiNote		note_event;
	GdamMidiSystemExclusive	sysex_event;
};

char* gdam_midi_event_get_string(GdamMidiEvent* event);

struct _GdamMidiDeviceClass {
	GtkObjectClass		object_class;

	/* Called on every midi event. */
	void                    (*event)      (GdamMidiDevice* midi_dev,
	                                       GdamMidiEvent*  event);

	void                    (*new_trap)   (GdamMidiDevice* midi_dev,
	                                       GdamMidiTrap*   trap);

	/* Someday deprecated... (i wish i could use gdam_midi_device_connect
	 *                        for these...) */
	void			(*note_on)    (GdamMidiDevice* midi_dev,
	                                       GdamMidiNote*   note);
	void			(*note_off)   (GdamMidiDevice* midi_dev,
	                                       GdamMidiNote*   note);
	/* Deprecated... 
	 * use ``gdam_midi_device_connect'' to listen for a particular
	 * event, or the ``event'' signal sometimes, for the bold
	 * or wastefu... */
	void			(*pitch_wheel)(GdamMidiDevice* midi_dev,
	                                       GdamMidiPitch*  pitch);
	void			(*midi_param) (GdamMidiDevice* midi_dev,
	                                       GdamMidiParam*  pitch);

	/* This function is passed in from derivers... */
        gboolean                (*parse_raw)  (GIOChannel*     io_channel,
                                               GIOCondition    condition,
					       gpointer        user_data);

	int			(*open_fd)    (GdamMidiDevice* device);
};

/*
 *                         GdamMidiTrap
 *
 * This should correspond to a particular knob or slider usually,
 * but in general it just means a certain type of midi event,
 * down to channel/parameter.
 */
struct _GdamMidiTrap {
	/* These four members uniquely identify a GdamMidiTrap. */
	GdamMidiEventType	type;
	int			channel;
	int			which_param;
	GdamMidiDevice*		device;

	/* if type == GDAM_MIDI_SYSEX, 
	   then
		channel == which_param == 0,
	   and these members uniquely identify the trap: */
	int                     sysex_length;
	guint8*			sysex_message;
	const char*             make_name;
	const char*             model_name;

	/* A helpful name for the trap. */
	char*			trap_name;

	int			min_value, max_value;
	GSList*			hooks;

};

guint gdam_midi_trap_hash(gconstpointer);
gint  gdam_midi_trap_equal(gconstpointer,gconstpointer);
char* gdam_midi_trap_get_string(GdamMidiTrap* trap);


/*=====================================================================*
 *                            GdamMidiDevice                           *
 *                                                                     *
 * A device, in the sense it is a single /dev/midi* file.              *
 *=====================================================================*/
struct _GdamMidiDevice {
	GtkObject		object;

	/* The device name. */
	char*			filename;

	/* A io channel opened to the device. */
	GIOChannel*		io_channel;

	/* Iff we are in the middle of a system exclusive
	 * then this is non-NULL. */
	GString*		system_exclusive;

	/* A way to lookup GdamMidiTraps */
	GHashTable*		trap_table;

	/* A way to look up the hooks by id. */
	GHashTable*		hook_id_table;

	/* Last hook_id we allocated (0 initially). */
	guint			last_hook_used;

	/* List of patch files (by filename) we've loaded */
	GSList*			patch_list;


	/* MIDI make and model id numbers. */
	int			make_id;
	int			model_id;

	int			triplet_length;
	unsigned char		triplet[3];
};

/* We maintain a higher level cache of all opened midi devices.
 * Get one using this function. */
GdamMidiDevice* gdam_midi_device_get_unique(const char* device_name);


/* Raw device open/close, for bypassing the caching system. */
/* Create / open midi devices. */
GdamMidiDevice* gdam_midi_device_new        ();
gboolean        gdam_midi_device_open       (GdamMidiDevice* midi_device,
                                             const char*     device_name);
void            gdam_midi_device_close      (GdamMidiDevice* midi_device);

/*
 * Load a `map file' for the midi device.
 * (We made this format up ourselves:
 *     pitchwheel 0 name
 *     parameter 0 0 name
 *     sysex length ... ... ... name
 *     etc.)
 */
gboolean        gdam_midi_device_load_map   (GdamMidiDevice*    midi_dev,
                                             const char*        db_filename);

guint           gdam_midi_device_connect    (GdamMidiDevice*    midi_dev,
				             GdamMidiEventType  type,
				             int                channel,
				             int                which_param,
				             const char*        hook_name,
				             GdamMidiHookFunc   hook_func,
				             GDestroyNotify     destroy,
				             gpointer           user_data,
				             GtkObject*         alive_object);
guint           gdam_midi_device_connect_sysex
                                            (GdamMidiDevice*    midi_dev,
				             int                length,
				             const guint8*      data,
				             const char*        hook_name,
				             GdamMidiHookFunc   hook_func,
				             GDestroyNotify     destroy,
				             gpointer           user_data,
				             GtkObject*         alive_object);
void            gdam_midi_device_disconnect (GdamMidiDevice*    midi_dev,
				             guint              hook_id);

/* If any of these parameters are NULL, then hooks are disconnected
 * with disregard for that type. (even midi_dev)
 *
 * Also: object == alive_object, hook_name == description.
 *
 * Yes, this function sucks and i know it... */
void            gdam_midi_device_disconnect_by_type
                                            (GdamMidiDevice*    midi_dev,
					     GtkObject*         object,
					     const char*        hook_name);


GdamMidiTrap*   gdam_midi_device_get_trap   (GdamMidiDevice*    midi_dev,
                                             GdamMidiEvent*     event,
					     gboolean           create);

/* Create a trap with a given name.  (Private) */
void            gdam_midi_device_create_trap(GdamMidiDevice*    midi_dev,
                                             GdamMidiEventType  type,
					     int                channel,
					     int                param,
					     const char*        trap_name);

void            gdam_midi_device_iterate_traps
                                            (GdamMidiDevice*    device,
					     GdamMidiTrapIterator iter_func,
					     gpointer           iter_data);

/* Create/destroy midi notes (mostly internal use). */
GdamMidiNote*	gdam_midi_note_new          (int                instrument,
                                             int		midi_id,
                                             int                velocity);
void            gdam_midi_note_destroy      (GdamMidiNote*      note);

/* For derivers, or those who want to simulate device inputs... */
void gdam_midi_device_emit_program          (GdamMidiDevice* device,
                                             int             channel,
				             int             program);
void gdam_midi_device_emit_channel_pressure (GdamMidiDevice* device,
                                             int             channel,
					     int             pressure);
void gdam_midi_device_emit_note_off         (GdamMidiDevice* device,
					     int             instrument,
                                             int             pitch,
					     int             volume);
void gdam_midi_device_emit_note_on          (GdamMidiDevice* device,
					     int             instrument,
                                             int             pitch,
					     int             volume);
void gdam_midi_device_emit_key_pressure     (GdamMidiDevice* device,
					     int             instrument,
                                             int             pitch,
					     int             volume);
void gdam_midi_device_emit_parameter        (GdamMidiDevice* device,
                                             int             channel,
					     int             which_param,
					     int             value);
void gdam_midi_device_emit_pitch_wheel      (GdamMidiDevice* device,
                                             int             channel,
					     int             arg);
void gdam_midi_device_emit_sysex            (GdamMidiDevice* device,
				             const char*     maker,
				             const char*     model,
                                             int             length,
				             const guint8*   bytes);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
