/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamskin.h"
#include "gdampacker.h"
#include <gdam/component/gdamutil.h>
#include "string.h"
#include "gdamkeyboard.h"
#include "gdamskinutil.h"
#include <glade/glade.h>
#include <gdam/component/gdamfilter.h>
#include <gdam/component/gdamturntable.h>
#include <gdam/component/gdambuilder.h>
#include <stdlib.h>
#include <ctype.h>

static void     gdam_packer_post_connecting_handler (GdamSkin*          skin);
static gboolean gdam_packer_parse_xml               (GdamView*          view,
	                                             xmlNode*           xml,
						     GdamXmlInfo*       info);
static gboolean gdam_packer_make_xml                (GdamView*          view,
	                                             xmlNode*           parent,
				                     GdamXmlInfo*       info);

static void gdam_packer_toplevel_connect(GdamSkin* skin, GdamToplevel* top)
{
	GdamPacker* packer = GDAM_PACKER(skin);
	packer->toplevel = top;
}

static void connect_widget_to_press(GtkWidget* widget)
{
	gdam_keyboard_connect_widget(widget, GDAM_KEYBOARD_CONNECT_PRESS);
}

static void gdam_packer_post_connecting_handler(GdamSkin* skin)
{
	GdamPacker* packer = GDAM_PACKER(skin);
	g_return_if_fail(packer != NULL);

	/* This assertion is so that no additional deferred handlers
	 * will be added; should be done by gdam_skin_new() */
	g_return_if_fail(skin->finished_connecting);

	if (packer->had_keyboard_handler) {
		packer->key_widget = skin->main_widget;
		if (GTK_WIDGET_REALIZED(packer->key_widget)) {
			connect_widget_to_press(packer->key_widget);
		} else {
			gtk_signal_connect
				(GTK_OBJECT(packer->key_widget),
				 "realize",
				 GTK_SIGNAL_FUNC(connect_widget_to_press),
				 NULL);
		}
	}
		
}

/*
         gdam_packer_set_filter_child:filter_name:child_name
    		Attach a GdamSkin/GdamModel as the child of
		a GdamSkin/GdamFilter.
*/

typedef void (*ConnectChildFunc)(GdamModel* parent, GdamModel* child);
typedef GdamModel* (*GetChildFunc)(GdamModel* parent);

typedef struct _DeferredFilterSetChild DeferredFilterSetChild;
struct _DeferredFilterSetChild {
	char* child_name;
	char* filter_name;
	ConnectChildFunc connect_func;
	GetChildFunc get_func;
};

static gboolean
do_set_filter_child_packing(GdamSkin* skin, gpointer user_data)
{
	DeferredFilterSetChild* set_child_info = user_data;
	gpointer child_ptr;
	gpointer filter_ptr;
	GdamModel* child_model;
	GdamModel* filter_model;
	char* child_name = set_child_info->child_name;
	char* filter_name = set_child_info->filter_name;
	GdamPacker* packer = GDAM_PACKER(skin);
	filter_ptr = g_hash_table_lookup(packer->named_skins, filter_name);
	child_ptr  = g_hash_table_lookup(packer->named_skins, child_name);
	if (filter_ptr == NULL) {
		g_warning("set_filter_child: %s not found", filter_name);
		return FALSE;
	}
	if (GDAM_IS_MODEL(filter_ptr))
		filter_model = GDAM_MODEL(filter_ptr);
	else
		filter_model = GDAM_SKIN(filter_ptr)->model;
	if (child_ptr == NULL) {
	 	child_model = (*set_child_info->get_func)(filter_model);
		if (child_model == NULL) {
			g_warning("set_filter_child: %s not found", child_name);
			return FALSE;
		}
		{
			GdamSkin* child_skin;
			child_skin = gdam_skin_find_from_model(child_model);
			if (child_skin != NULL)
				child_ptr = child_skin;
			else
				child_ptr = child_model;
			g_hash_table_insert(packer->named_skins,
			                    g_strdup(child_name),
					    child_ptr);
		}
		return TRUE;
	}
	if (GDAM_IS_MODEL(child_ptr))
		child_model = GDAM_MODEL(child_ptr);
	else
		child_model = GDAM_SKIN(child_ptr)->model;
	g_return_val_if_fail(filter_model != NULL, FALSE);
	g_return_val_if_fail(child_model != NULL, FALSE);
	(*set_child_info->connect_func)(filter_model, child_model);
	return TRUE;
}

static void free_set_filter_child_info(gpointer user_data)
{
	DeferredFilterSetChild* set_child_info = user_data;
	g_free(set_child_info->child_name);
	g_free(set_child_info->filter_name);
	g_free(set_child_info);
}

static gboolean gdam_packer_set_filter_child_handler
				(GdamSkin*          skin,
			         GtkObject*         object,
			         const char*        signal_name,
			         int                num_args,
			         char**             command)
{
	char* filter_name;
	char* child_name;
	DeferredFilterSetChild* set_child_info;
	if ((filter_name = command[1]) == NULL
	 || (child_name = command[2]) == NULL) {
		g_warning("too few arguments to %s", command[0]);
		return TRUE;
	}
	/* NULL the strings out so `cleanup_strings' doesn't
	 * free them. */
	command[1] = command[2] = NULL;

	set_child_info = g_new0(DeferredFilterSetChild, 1);
	set_child_info->child_name = child_name;
	set_child_info->filter_name = filter_name;
	set_child_info->connect_func = (ConnectChildFunc)gdam_filter_set_child;
	set_child_info->get_func = (GetChildFunc)gdam_filter_get_child;

	gdam_skin_add_initializer(skin,
	                          do_set_filter_child_packing,
				  free_set_filter_child_info,
				  set_child_info);
	return TRUE;
}

static gboolean gdam_packer_set_builder_child_handler
				(GdamSkin*          skin,
			         GtkObject*         object,
			         const char*        signal_name,
			         int                num_args,
			         char**             command)
{
	char* builder_name;
	char* child_name;
	DeferredFilterSetChild* set_child_info;
	if ((builder_name = command[1]) == NULL
	 || (child_name = command[2]) == NULL) {
		g_warning("too few arguments to %s", command[0]);
		return TRUE;
	}
	/* NULL the strings out so `cleanup_strings' doesn't
	 * free them. */
	command[1] = command[2] = NULL;

	set_child_info = g_new0(DeferredFilterSetChild, 1);
	set_child_info->child_name = child_name;
	set_child_info->filter_name = builder_name;
	set_child_info->connect_func = (ConnectChildFunc)gdam_builder_set_component;
	set_child_info->get_func = (GetChildFunc)gdam_builder_get_component;

	gdam_skin_add_initializer(skin,
	                          do_set_filter_child_packing,
				  free_set_filter_child_info,
				  set_child_info);
	return TRUE;
}

static gboolean gdam_packer_set_turntable_child_handler
				(GdamSkin*          skin,
			         GtkObject*         object,
			         const char*        signal_name,
			         int                num_args,
			         char**             command)
{
	char* turntable_name;
	char* child_name;
	DeferredFilterSetChild* set_child_info;
	if ((turntable_name = command[1]) == NULL
	 || (child_name = command[2]) == NULL) {
		g_warning("too few arguments to %s", command[0]);
		return TRUE;
	}

	/* NULL the strings out so `cleanup_strings' doesn't
	 * free them. */
	command[1] = command[2] = NULL;

	set_child_info = g_new0(DeferredFilterSetChild, 1);
	set_child_info->child_name = child_name;
	set_child_info->filter_name = turntable_name;
	set_child_info->connect_func = 
				(ConnectChildFunc)gdam_turntable_set_producer;
	set_child_info->get_func = 
				(GetChildFunc)gdam_turntable_get_producer;

	gdam_skin_add_initializer(skin,
	                          do_set_filter_child_packing,
				  free_set_filter_child_info,
				  set_child_info);
	return TRUE;
}

/*
         gdam_packer_add_toplevel:skin_name
*/
typedef struct _PackerToplevelInfo PackerToplevelInfo;
struct _PackerToplevelInfo {
	char* name;
	gboolean misses_data;
};
static void packer_toplevel_info_free(gpointer ptr)
{
	PackerToplevelInfo* info = ptr;
	g_free(info->name);
	g_free(info);
}
static gboolean do_toplevel_packer(GdamSkin* skin, gpointer user_data)
{
	PackerToplevelInfo* info = user_data;
	char* skin_name = info->name;
	GdamPacker* packer = GDAM_PACKER(skin);
	GdamSkin* subskin;
	GdamModel* model;

	subskin = g_hash_table_lookup(packer->named_skins, skin_name);
	if (subskin == NULL) {
		g_warning("no skin named `%s': cannot add it to toplevel",
			  skin_name);
		return TRUE;
	}
	if ((model=subskin->model) == NULL) {
		g_warning("skin `%s' didn't have model: "
		          "cannot add to toplevel", skin_name);
		return TRUE;
	}
	if (packer->toplevel == NULL) {
		return FALSE; 
	}

	gdam_toplevel_add_child(packer->toplevel, model);
	gdam_toplevel_set_misses_data(packer->toplevel,
				      model,
				      info->misses_data);
	return TRUE;
}
static gboolean gdam_packer_add_toplevel_handler
				(GdamSkin*          skin,
			         GtkObject*         object,
			         const char*        signal_name,
			         int                num_args,
			         char**             command)
{
	char* child_name;
	PackerToplevelInfo* info;
	gboolean misses_data = TRUE;

	if ((child_name = command[1]) == NULL) {
		g_warning("too few arguments to %s", command[0]);
		return TRUE;
	}

	/* NULL the strings out so `cleanup_strings' doesn't
	 * free them. */
	command[1] = NULL;
	if (command[2] != NULL
	 && strcmp(command[2], "synchronized") == 0) {
		misses_data = FALSE;
	}

	info = g_new(PackerToplevelInfo, 1);
	info->name = child_name;
	info->misses_data = misses_data;

	gdam_skin_add_initializer(skin,
	                          do_toplevel_packer,
				  packer_toplevel_info_free,
				  info);
	return TRUE;
}


/*
         gdam_packer_create:local_name:skin_type:skin_filename:options
    		Create a skin as an element of this container.
		 (Usually a container for just one widget)
*/
static gboolean gdam_packer_create_handler
				(GdamSkin*          skin,
			         GtkObject*         object,
			         const char*        signal_name,
			         int                num_args,
			         char**             command)
{
	char* local_name;
	char* skin_type;
	char* skin_filename;
	GtkContainer* container = GTK_CONTAINER(object);
	GdamPacker* packer;
	GdamSkin* sub_skin;
	GdamModel* sub_model;
	GtkWidget* main_widget;
	g_return_val_if_fail(container != NULL, TRUE);

	if ((local_name    = command[1]) == NULL
	 || (skin_type     = command[2]) == NULL
	 || (skin_filename = command[3]) == NULL) {
		g_warning("not enough arguments to `packer_create'");
		return TRUE;
	}

	packer = GDAM_PACKER(skin);
	g_return_val_if_fail(packer != NULL, TRUE);

	sub_skin = g_hash_table_lookup(packer->named_skins, local_name);
	if (sub_skin != NULL) {
		gdam_debug("not creating named skin %s, already exists",
			local_name);
		g_return_val_if_fail(GDAM_IS_SKIN(packer), TRUE);
		sub_model = sub_skin->model;
	} else {
		sub_skin = gdam_skin_new_by_type_name(skin_type, 
						     skin_filename,
						     num_args - 4,
						     command + 4,
						     skin->channel,
						     &sub_model);
		/* Set it to NULL to it isn't freed... */
		command[1] = NULL;
		g_hash_table_insert(packer->named_skins, local_name, sub_skin);
	}
	g_return_val_if_fail(sub_skin != NULL, TRUE);

	if (sub_model && packer->toplevel) {
		gdam_toplevel_add_child(packer->toplevel, sub_model);
	}

	main_widget = sub_skin->main_widget;
	g_return_val_if_fail(main_widget != NULL, TRUE);
	gdam_set_container(container, main_widget);
	gtk_widget_show(main_widget);

	/* Not sure if this assertion is really needed,
	 * but it seems like a lot of later code is
	 * ambiguous otherwise... */
	g_return_val_if_fail(sub_skin->model == sub_model, TRUE);

	return TRUE;
}
/*
	   gdam_packer_trap_key_press:key:skin:....

	   Connects a keypress like a button or something... */
typedef struct _DeferredKeyTrap DeferredKeyTrap;
struct _DeferredKeyTrap {
	char*			subskin_name;
	int                     keycode;
	gboolean                is_press;
	int                     num_args;
	char**                  args;
};
static gboolean deferred_key_trap_do_trap(GdamSkin* skin, gpointer user_data)
{
	GdamSkin* subskin;
	DeferredKeyTrap* def_trap = user_data;
	GdamKeyboardTrap* trap;
	GdamPacker* packer = GDAM_PACKER(skin);
	subskin = gdam_packer_get_skin(packer, def_trap->subskin_name);
	g_return_val_if_fail(subskin != NULL, FALSE);
	trap = gdam_keyboard_get_trap(gdam_keyboard_get_instance(),
	                              packer->key_widget,
				      def_trap->keycode);
	if (!gdam_skin_do_connect(subskin,
	                          GTK_OBJECT(trap),
				  "pressed",
				  def_trap->num_args,
				  def_trap->args)) {
		g_warning("gdam_skin_do_connect failed trapping key event: "
		          "num_args=%d, handler_hame=%s, skintype=%s",
			  def_trap->num_args,
			  def_trap->args[0],
			  gtk_type_name(GTK_OBJECT(subskin)->klass->type));
		return TRUE;
	}
	gdam_debug("gdam_skin_do_connect succeeded trapping key event: "
		  "num_args=%d, handler_hame=%s, skintype=%s, keycode=%08x",
		  def_trap->num_args,
		  def_trap->args[0],
		  gtk_type_name(GTK_OBJECT(subskin)->klass->type),
		  def_trap->keycode);
	return TRUE;
}
static void deferred_key_trap_free(gpointer user_data)
{
	DeferredKeyTrap* def_trap = user_data;
	int i;
	for (i = 0; i < def_trap->num_args; i++)
		g_free(def_trap->args[i]);
	g_free(def_trap->args);
	g_free(def_trap);
}

static guint parse_key_code(const char* keyname)
{
	g_return_val_if_fail(keyname != NULL, 0);
	if (keyname[0] == '\0') {
		g_warning("empty keycode encountered");
		return 0;
	}
	if (keyname[1] == '\0')
		return (guint) (*(unsigned char*)keyname);
	if (keyname[0] == '^' && keyname[2] == '\0') {
		int key = tolower(keyname[1]);
		return key | GDAM_KEYBOARD_CTRL_MASK;
	}
	if (strncasecmp(keyname, "ctrl-", 5) == 0
	 || strncasecmp(keyname, "control-", 8) == 0) {
		int key = (unsigned char) (strchr(keyname, '-')[1]);
		return key | GDAM_KEYBOARD_CTRL_MASK;
	}
	if (strncasecmp(keyname, "alt-", 4) == 0) {
		int key = (unsigned char) (strchr(keyname, '-')[1]);
		return key | GDAM_KEYBOARD_ALT_MASK;
	}

	{
		guint key = gdk_keyval_from_name(keyname);
		if (key != 0)
			return key;
	}

	g_warning("cannot parse key code string: %s", keyname);
	return 0;
}

static gboolean gdam_packer_trap_key_handler
				(GdamSkin*          skin,
			         GtkObject*         object,
			         const char*        signal_name,
			         int                num_args,
			         char**             command)
{
	GdamPacker* packer = GDAM_PACKER(skin);
	DeferredKeyTrap* deferred_trap;
	char* subskin_name;
	int key_code;
	if (num_args < 3) {
		g_warning("need 3 args to trap_key_press");
		return TRUE;
	}
	key_code = parse_key_code(command[1]);
	if (key_code == 0) {
		g_warning("error parsing key code: %s", command[1]);
		return TRUE;
	}
	subskin_name = command[2];
	command[2] = NULL;
	packer->had_keyboard_handler = TRUE;
	deferred_trap = g_new(DeferredKeyTrap, 1);
	deferred_trap->subskin_name = subskin_name;
	deferred_trap->keycode = key_code;
	deferred_trap->is_press
		= strcmp("gdam_packer_trap_key_release", command[0]) == 0
							? FALSE : TRUE;
	deferred_trap->num_args = num_args - 3;
	deferred_trap->args = gdam_strdupv(command + 3);
	gdam_skin_add_initializer(GDAM_SKIN(packer),
                                  deferred_key_trap_do_trap,
                                  deferred_key_trap_free,
				  deferred_trap);
	return TRUE;
}


static GdamSkinGenericConnect connect_infos[] =
{
	{ "gdam_packer_set_filter_child",
	   gdam_packer_set_filter_child_handler },
	{ "gdam_packer_set_builder_child",
	   gdam_packer_set_builder_child_handler},
	{ "gdam_packer_set_turntable_child",
	   gdam_packer_set_turntable_child_handler},
        { "gdam_packer_add_toplevel",
           gdam_packer_add_toplevel_handler	},
	{ "gdam_packer_trap_key_press",
	   gdam_packer_trap_key_handler   	},
	{ "gdam_packer_trap_key_release",
	   gdam_packer_trap_key_handler   	},
	{ "gdam_packer_create",
	   gdam_packer_create_handler           }		};


static void gdam_packer_init(GdamPacker* packer)
{
	packer->named_skins = g_hash_table_new(g_str_hash, g_str_equal);
}



static void gdam_packer_class_init(GdamPackerClass* packer_class)
{
	GdamSkinClass* skin_class = GDAM_SKIN_CLASS(packer_class);
	GdamViewClass* view_class;
	view_class = GDAM_VIEW_CLASS(packer_class);
	skin_class->post_connecting_handler
					= gdam_packer_post_connecting_handler;
	skin_class->toplevel_connect = gdam_packer_toplevel_connect;
	view_class->parse_xml = gdam_packer_parse_xml;
	view_class->make_xml = gdam_packer_make_xml;
	gdam_skin_class_register_generic_handlers(skin_class,
				                  G_N_ELEMENTS(connect_infos),
				                  connect_infos);
}

GtkType gdam_packer_get_type()
{
	static GtkType packer_type = 0;
	if (!packer_type) {
		static const GtkTypeInfo packer_info = {
			"GdamPacker",
			sizeof(GdamPacker),
			sizeof(GdamPackerClass),
			(GtkClassInitFunc) gdam_packer_class_init,
			(GtkObjectInitFunc) gdam_packer_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_SKIN;
		packer_type = gtk_type_unique(parent, &packer_info);
	}
	return packer_type;
}

GdamSkinClass* gdam_packer_class()
{
	static GdamSkinClass* class = NULL;
	if (class != NULL)
		return class;
	return class = GDAM_SKIN_CLASS(gtk_type_class(GDAM_TYPE_PACKER));
}

#if 0
typedef struct _DeferredEvent DeferredEvent;

struct _DeferredEvent {
	GdamPackerFunc      command;
	GDestroyNotify      done_with_command;
	gpointer            user_data;
};

static void run_and_delete_one(DeferredEvent* event, GdamPacker* packer)
{
	(*event->command)(packer, event->user_data);
	if (event->done_with_command)
		(*event->done_with_command)(event->user_data);
	g_free(event);
}

static void gdam_packer_run_deferred(GdamPacker* packer)
{
	GSList* deferred = packer->pending;
	packer->pending = NULL;
	deferred = g_slist_reverse(deferred);
	g_slist_foreach(deferred, (GFunc) run_and_delete_one, packer);
	g_slist_free(deferred);
}

void        gdam_packer_add_deferred_command(GdamPacker*         packer,
                                             GdamPackerFunc      command,
			                     GDestroyNotify      destroyer,
			                     gpointer            user_data)
{
	DeferredEvent* event;
	event = g_new(DeferredEvent, 1);
	event->command           = command;
	event->done_with_command = destroyer;
	event->user_data         = user_data;
	packer->pending = g_slist_prepend(packer->pending, event);
}
#endif

GdamModel* gdam_packer_get_model               (GdamPacker*         packer,
                                                const char*         name)
{
	gpointer ptr = g_hash_table_lookup(packer->named_skins, name);
	g_return_val_if_fail(ptr != NULL, NULL);
	if (GDAM_IS_MODEL(ptr))
		return GDAM_MODEL(ptr);
	else {
		GdamSkin* skin;
		skin = GDAM_SKIN(ptr);
		return skin->model;
	}
}

GdamSkin*      gdam_packer_get_skin            (GdamPacker*         packer,
			                        const char*         name)
{
	gpointer ptr = g_hash_table_lookup(packer->named_skins, name);
	g_return_val_if_fail(ptr != NULL, NULL);
	if (GDAM_IS_MODEL(ptr)) {
		g_warning("%s was a component, not a skin", name);
		return NULL;
	} else {
		GdamSkin* skin;
		skin = GDAM_SKIN(ptr);
		return skin;
	}
}

/*
 * Parse/produce xml file.
 */

typedef struct _XmlMaker XmlMaker;
struct _XmlMaker {
	GdamPacker*		packer;
	xmlNode*		parent;
	GdamXmlInfo*            xml_info;
};

static void add_one_xml_skin(const char*        name,
                             gpointer           model_or_skin,
			     XmlMaker*          maker)
{
	xmlNode* obj_node = NULL;
	xmlNode* packer_node;
	gboolean was_skin = FALSE;
	int id;
	g_return_if_fail(model_or_skin != NULL);

	id = gdam_xml_info_lookup_by_object(maker->xml_info, model_or_skin);
	if (id != 0) {
		char id_buf[256];
		sprintf(id_buf, "%d", id);
		was_skin = GDAM_IS_VIEW(model_or_skin);
		packer_node = xmlNewNode(NULL, was_skin ? "skin" : "component");
		xmlNewProp(packer_node, "name", name);
		xmlAddChild(maker->parent, packer_node);
		xmlNewChild(packer_node, NULL, "id", id_buf);
		return;
	} else if (GDAM_IS_MODEL(model_or_skin)) {
		GdamModel* comp = GDAM_MODEL(model_or_skin);
		obj_node = gdam_model_to_xml(comp, maker->xml_info);
		was_skin = FALSE;
	} else if (GDAM_IS_SKIN(model_or_skin)) {
		GdamSkin* skin = GDAM_SKIN(model_or_skin);
		obj_node = gdam_skin_to_xml(skin, maker->xml_info);
		was_skin = TRUE;
	} else if (GTK_IS_OBJECT(model_or_skin)) {
		g_warning("%p [%s] is not a skin or component", model_or_skin,
			gtk_type_name(GTK_OBJECT(model_or_skin)->klass->type));
		return;
	} else {
		g_warning("gdam_packer: "
			"add_one_xml: %p is not a gtkobject", model_or_skin);
		return;
	}
	if (obj_node == NULL) {
		g_warning("error building xml [in packer]...");
		return;
	}

	packer_node = xmlNewNode(NULL, was_skin ? "skin" : "component");
	xmlNewProp(packer_node, "name", name);
	xmlAddChild(packer_node, obj_node);
	xmlAddChild(maker->parent, packer_node);
}

static gboolean gdam_packer_make_xml(GdamView* skin,
	                             xmlNode*           parent,
				     GdamXmlInfo*       xml_info)
{
	GdamPacker* packer = GDAM_PACKER(skin);
	if (packer->named_skins != NULL) {
		XmlMaker xml_maker;
		xml_maker.packer   = packer;
		xml_maker.parent   = parent;
		xml_maker.xml_info = xml_info;
		g_hash_table_foreach(packer->named_skins,
		                     (GHFunc) add_one_xml_skin,
				     &xml_maker);
	}

	/* XXX: this should be inherited??? */
	gdam_skin_add_filename_xml(GDAM_SKIN(skin), parent);
		                     
	/* XXX: store interconnect ??? */
	return TRUE;
}

static gboolean gdam_packer_parse_xml(GdamView*      view,
	                              xmlNode*       xml,
				      GdamXmlInfo*   xml_info)
{
	GdamPacker* packer = GDAM_PACKER(view);
	GdamChannel* channel;
	xmlNode* node;
	channel = GDAM_SKIN(packer)->channel;

	/* XXX: get rid of this somehow... */
	packer->toplevel = gdam_toplevel_get_global(channel);

	for (node = xml; node != NULL; node = node->next) {
		if (strcmp(node->name, "filename") == 0) {
			gdam_skin_handle_filename_xml(GDAM_SKIN(packer), node);
			continue;
		}
		
		if (strcmp(node->name, "skin") == 0
		 || strcmp(node->name, "component") == 0) {
			char* name;
			GtkObject* obj;
			xmlNode* child;


			name = xmlGetProp(node, "name");
			if (name == NULL) {
				g_warning("<%s> must have name in packer",
				            node->name);
				continue;
			}
			child = node->childs;
			if (child == NULL || child->next != NULL) {
				g_warning("<%s> must contain exactly one "
					"subnode", node->name);
				continue;
			}

			if (strcmp(node->name, "skin") == 0) {
				GdamSkin* skin;
				skin = gdam_skin_from_xml(child,
				                          channel,
							  xml_info);
				obj = GTK_OBJECT(skin);
			} else {
				GdamModel* component;
				component = gdam_model_from_xml(child,
				                                channel,
								xml_info);
				obj = GTK_OBJECT(component);
			}
			if (obj != NULL) {
				g_hash_table_insert(packer->named_skins,
						    name,
						    obj);
			}
			continue;
		}
		g_warning("unhandled node type in packer: <%s>", node->content);
	}
	return TRUE;
}
